from django.urls import path

from response.slack import views

urlpatterns = [
    path("slash_command", views.slash_command, name="slash_command"),
    path("action", views.action, name="action"),
    path("event", views.event, name="event"),
    path("cron_minute", views.cron_minute, name="cron_minute"),
    path("cron_daily", views.cron_daily, name="cron_daily"),
]



import datetime
import unittest
from unittest.mock import patch

import pytest
from faker import Faker

from response.core.models import TimelineEvent
from response.slack.event_handlers import handle_pin_added, handle_pin_removed
from response.slack.models import PinnedMessage
from tests.factories import ExternalUserFactory, IncidentFactory

faker = Faker()


class TestPinnedMessage(unittest.TestCase):
    @pytest.mark.django_db
    @patch("response.slack.models.pinned_message.get_user_profile")
    def test_add_and_remove_pin(self, get_user_profile):
        incident = IncidentFactory.create()
        user = ExternalUserFactory.create()

        get_user_profile.return_value = {"name": user.display_name}

        text = faker.paragraph(nb_sentences=2, variable_nb_sentences=True)
        handle_pin_added(
            incident,
            {"item": {"message": {"user": user.external_id, "ts": 123, "text": text}}},
        )
        get_user_profile.assert_called_with(user.external_id)

        message = PinnedMessage.objects.get(incident=incident, message_ts=123)
        assert message.text == text

        handle_pin_removed(
            incident,
            {"item": {"message": {"user": user.external_id, "ts": 123, "text": text}}},
        )

        get_user_profile.assert_called_with(user.external_id)

        with pytest.raises(PinnedMessage.DoesNotExist):
            PinnedMessage.objects.get(incident=incident, message_ts=123)
            TimelineEvent.objects.get(
                incident=incident, timestamp=datetime.fromtimestamp(123)
            )



from datetime import datetime

from django.db import models

from response.core.models import ExternalUser, Incident
from response.slack.cache import get_user_profile


class UserStats(models.Model):
    user = models.ForeignKey(
        ExternalUser, on_delete=models.CASCADE, blank=False, null=False
    )
    incident = models.ForeignKey(
        Incident, on_delete=models.CASCADE, blank=False, null=False
    )

    join_time = models.DateTimeField(null=True)
    message_count = models.IntegerField(default=0)

    class Meta:
        unique_together = ("incident", "user")

    @staticmethod
    def increment_message_count(incident, user_id):
        name = get_user_profile(user_id)["name"]
        user, _ = ExternalUser.objects.get_or_create_slack(
            external_id=user_id, display_name=name
        )

        user_stats, created = UserStats.objects.get_or_create(
            incident=incident, user=user
        )

        if created:
            user_stats.join_time = datetime.now()

        user_stats.message_count += 1
        user_stats.save()

    def __str__(self):
        return f"{self.user.display_name} - {self.incident}"



from datetime import datetime

from django.db import models

from response import core, slack
from response.core.models.user_external import ExternalUser
from response.core.util import sanitize


class IncidentManager(models.Manager):
    def create_incident(
        self,
        report,
        reporter,
        report_time,
        report_only,
        private=False,
        summary=None,
        impact=None,
        lead=None,
        severity=None,
    ):
        incident = self.create(
            report=report,
            reporter=reporter,
            report_time=report_time,
            report_only=report_only,
            private=private,
            start_time=report_time,
            summary=summary,
            impact=impact,
            lead=lead,
            severity=severity,
        )
        return incident


class Incident(models.Model):

    objects = IncidentManager()

    # Reporting info
    report = models.CharField(max_length=200)
    reporter = models.ForeignKey(
        ExternalUser,
        related_name="reporter",
        on_delete=models.PROTECT,
        blank=False,
        null=True,
    )
    report_time = models.DateTimeField()
    report_only = models.BooleanField(default=False)
    private = models.BooleanField(default=False)

    start_time = models.DateTimeField(null=False)
    end_time = models.DateTimeField(blank=True, null=True)

    # Additional info
    summary = models.TextField(
        blank=True, null=True, help_text="What's the high level summary?"
    )
    impact = models.TextField(
        blank=True, null=True, help_text="What impact is this having?"
    )
    lead = models.ForeignKey(
        ExternalUser,
        related_name="lead",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        help_text="Who is leading?",
    )

    # Severity
    SEVERITIES = (("1", "critical"), ("2", "major"), ("3", "minor"), ("4", "trivial"))
    severity = models.CharField(
        max_length=10, blank=True, null=True, choices=SEVERITIES
    )

    def __str__(self):
        return self.report

    def comms_channel(self):
        try:
            return slack.models.CommsChannel.objects.get(incident=self)
        except slack.models.CommsChannel.DoesNotExist:
            return None

    def duration(self):
        delta = (self.end_time or datetime.now()) - self.start_time

        hours, remainder = divmod(delta.total_seconds(), 3600)
        minutes, seconds = divmod(remainder, 60)

        time_str = ""
        if hours > 0:
            hours = int(hours)
            time_str += f"{hours} hrs " if hours > 1 else f"{hours} hr "

        if minutes > 0:
            minutes = int(minutes)
            time_str += f"{minutes} mins " if minutes > 1 else f"{minutes} min "

        if hours == 0 and minutes == 0:
            seconds = int(seconds)
            time_str += f"{seconds} secs"

        return time_str.strip()

    def is_closed(self):
        return True if self.end_time else False

    def severity_text(self):
        for sev_id, text in self.SEVERITIES:
            if sev_id == self.severity:
                return text
        return None

    def severity_emoji(self):
        if not self.severity:
            return "☁️"

        return {"1": "⛈️", "2": "🌧️", "3": "🌦️", "4": "🌤️"}[self.severity]

    def status_text(self):
        if self.report_only:
            return "reported"
        elif self.is_closed():
            return "resolved"
        else:
            return "live"

    def status_emoji(self):
        if self.report_only:
            return ":notebook:"
        elif self.is_closed():
            return ":droplet:"
        else:
            return ":fire:"

    def action_items(self):
        return core.models.Action.objects.filter(incident=self)

    def timeline_events(self):
        return core.models.TimelineEvent.objects.filter(incident=self)

    def save(self, *args, **kwargs):
        self.impact = sanitize(self.impact)
        self.report = sanitize(self.report)
        self.summary = sanitize(self.summary)
        super(Incident, self).save(*args, **kwargs)



import random

import factory
from faker import Factory

from response.core.models import Event

faker = Factory.create()


class EventFactory(factory.DjangoModelFactory):
    class Meta:
        model = Event

    timestamp = factory.LazyFunction(
        lambda: faker.date_time_between(start_date="-6m", end_date="now", tzinfo=None)
    )
    event_type = random.choice([Event.INCIDENT_EVENT_TYPE])

    # Using an Incident/Event factory here fails with a mysterious error:
    # https://github.com/pytest-dev/pytest-django/issues/713 (using @pytest.mark...
    # didn't resolve it). For now, a static fixture suffices.
    payload = {"report": "we are out of milk", "impact": "making tea is difficult"}



from django.conf import settings


class Dialog:
    def __init__(self, title, submit_label, elements=None, state=None):
        self.title = title
        self.submit_label = submit_label
        self.state = state
        self.elements = elements

    def add_element(self, element):
        if not self.elements:
            self.elements = []
        self.elements.append(element)

    def set_state(self, state):
        self.state = state

    def build_dialog(self, callback_id):
        elements = []
        for element in self.elements:
            elements.append(element.serialize())

        dialog = {
            "title": self.title,
            "submit_label": self.submit_label,
            "callback_id": callback_id,
            "state": self.state,
            "elements": elements,
        }
        return dialog

    def send_open_dialog(self, callback_id, trigger_id):
        """
        Open the dialog
        """
        return settings.SLACK_CLIENT.dialog_open(
            dialog=self.build_dialog(callback_id), trigger_id=trigger_id
        )


class Element:
    def __init__(self, label, name, optional, hint, subtype, value, placeholder):
        self.label = label
        self.name = name
        self.optional = optional
        self.hint = hint
        self.subtype = subtype
        self.value = value
        self.placeholder = placeholder

    def serialize(self):
        return {k: v for k, v in vars(self).items() if v}


class Text(Element):
    def __init__(
        self,
        label=None,
        name=None,
        optional=False,
        hint=None,
        subtype=None,
        value=None,
        placeholder=None,
    ):
        super().__init__(label, name, optional, hint, subtype, value, placeholder)
        self.type = "text"


class TextArea(Element):
    def __init__(
        self,
        label=None,
        name=None,
        optional=False,
        hint=None,
        subtype=None,
        value=None,
        placeholder=None,
    ):
        super().__init__(label, name, optional, hint, subtype, value, placeholder)
        self.type = "textarea"


class SelectWithOptions(Element):
    def __init__(
        self,
        options,
        label=None,
        name=None,
        optional=False,
        hint=None,
        subtype=None,
        value=None,
        placeholder=None,
    ):
        super().__init__(label, name, optional, hint, subtype, value, placeholder)
        self.type = "select"
        self.options = [{"label": l, "value": v} for l, v in options]


class SelectFromUsers(Element):
    def __init__(
        self,
        label=None,
        name=None,
        optional=False,
        hint=None,
        subtype=None,
        value=None,
        placeholder=None,
    ):
        super().__init__(label, name, optional, hint, subtype, value, placeholder)
        self.type = "select"
        self.data_source = "users"



import logging

import after_response

logger = logging.getLogger(__name__)

DIALOG_HANDLERS = {}


def dialog_handler(callback_id, func=None):
    def _wrapper(fn):
        DIALOG_HANDLERS[callback_id] = fn

    if func:
        return _wrapper(func)
    return _wrapper


def remove_dialog_handler(callback_id):
    DIALOG_HANDLERS.pop(callback_id, None)


@after_response.enable
def handle_dialog(payload):
    callback_id = payload["callback_id"]
    if callback_id not in DIALOG_HANDLERS:
        logger.error(f"Can't find handler for dialog id {callback_id}")
        return

    callback = DIALOG_HANDLERS[callback_id]

    user_id = payload["user"]["id"]
    channel_id = payload["channel"]["id"]
    submission = payload["submission"]
    response_url = payload["response_url"]
    state = payload["state"]

    callback(user_id, channel_id, submission, response_url, state)



import os
from datetime import datetime
from unittest.mock import MagicMock
from urllib.parse import urlencode

import pytest
from django.conf import settings
from django.urls import reverse

from response.slack.authentication import generate_signature
from response.slack.client import SlackClient


@pytest.fixture(autouse=True)
def mock_slack(monkeypatch):
    mock_slack = MagicMock(spec=SlackClient(""))
    mock_slack.get_channel_name.return_value = "#inc-test-channel"
    mock_slack.send_or_update_message_block.return_value = {"ok": True, "ts": 1234}
    monkeypatch.setattr(settings, "SLACK_CLIENT", mock_slack)
    return mock_slack


@pytest.fixture(scope="session")
def slack_signing_secret():
    return os.getenv("SLACK_SIGNING_SECRET", "signingsecretnotset")


@pytest.fixture
def post_from_slack_api(client, slack_signing_secret):
    def _post(url_name, data):
        form = urlencode(data).encode()
        timestamp = str(int(datetime.now().timestamp()))
        signature = generate_signature(timestamp, slack_signing_secret, form)

        return client.post(
            reverse(url_name),
            data=form,
            content_type="application/x-www-form-urlencoded",
            HTTP_X_SLACK_REQUEST_TIMESTAMP=timestamp,
            HTTP_X_SLACK_SIGNATURE=signature,
        )

    return _post



from .core.models import Action, Event, ExternalUser, Incident, TimelineEvent
from .slack.models import (
    CommsChannel,
    HeadlinePost,
    Notification,
    PinnedMessage,
    UserStats,
)

__all__ = (
    "Action",
    "Event",
    "Incident",
    "TimelineEvent",
    "ExternalUser",
    "CommsChannel",
    "HeadlinePost",
    "Notification",
    "PinnedMessage",
    "UserStats",
)



from .action_handler import ActionContext, action_handler, handle_action
from .dialog_handler import dialog_handler, handle_dialog
from .event_handler import handle_event, slack_event
from .headline_post_action import headline_post_action
from .incident_command import handle_incident_command, incident_command
from .incident_notification import (
    handle_notifications,
    recurring_notification,
    single_notification,
)
from .keyword_handler import handle_keywords, keyword_handler

__all__ = (
    "ActionContext",
    "action_handler",
    "dialog_handler",
    "handle_event",
    "slack_event",
    "headline_post_action",
    "handle_incident_command",
    "incident_command",
    "recurring_notification",
    "single_notification",
    "handle_keywords",
    "keyword_handler",
    "handle_action",
    "handle_dialog",
    "handle_notifications",
)



from .comms_channel import CommsChannel
from .headline_post import HeadlinePost
from .notification import Notification
from .pinned_message import PinnedMessage
from .user_stats import UserStats

__all__ = ("CommsChannel", "HeadlinePost", "Notification", "PinnedMessage", "UserStats")



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0007_delete_incidentextension")]

    operations = [
        migrations.AddField(
            model_name="externaluser",
            name="email",
            field=models.CharField(blank=True, max_length=100, null=True),
        )
    ]



import bleach
import markdown2
from django import template
from django.conf import settings

register = template.Library()


@register.filter
def markdown_filter(text):
    text = markdown2.markdown(text)
    html = bleach.clean(
        text,
        tags=settings.MARKDOWN_FILTER_WHITELIST_TAGS,
        attributes=settings.MARKDOWN_FILTER_WHITELIST_ATTRIBUTES,
        styles=settings.MARKDOWN_FILTER_WHITELIST_STYLES,
    )
    return bleach.linkify(html)



import json

from django.db import models


class Event(models.Model):
    ACTION_EVENT_TYPE = "action_event"
    INCIDENT_EVENT_TYPE = "incident_event"

    timestamp = models.DateTimeField()
    event_type = models.CharField(max_length=50)
    payload = models.TextField()

    def save(self, *args, **kwargs):
        self.payload = json.dumps(self.payload)
        super().save(*args, **kwargs)



import logging

from response.core.models.incident import Incident
from response.slack.models.comms_channel import CommsChannel

logger = logging.getLogger(__name__)

KEYWORD_HANDLERS = {}


def keyword_handler(keywords, func=None):
    def _wrapper(fn):
        for keyword in keywords:
            KEYWORD_HANDLERS[keyword] = fn
        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def handle_keywords(incident: Incident, payload):
    text = payload.get("text", "")
    user = payload.get("user", "")
    ts = payload.get("ts", "")

    comms_channel = CommsChannel.objects.get(incident=incident)

    for keyword, handler in KEYWORD_HANDLERS.items():
        if keyword.lower() in text.lower():
            handler(comms_channel, user, keyword, text, ts)



"""
This is a slightly more complex migration. This change adds a new CommsChannel
field to cache the channel name, updating it when the Slack API notifies us of
a channel rename. For existing CommsChannel in the DB, we need to populate
the channel_name by connecting to the Slack API.

This applies the migration in three stages. First, we add the new field, but
make it nullable, so there's no need to set a default. The next step iterates
over existing CommsChannels, and refreshes the name from the Slack API.

Finally, now that every CommsChannel has a channel_name set, we alter the field
so that a value is required.
"""

from django.conf import settings
from django.db import OperationalError, migrations, models

from response.slack.client import SlackError


def set_comms_channel_names(apps, schema_editor):
    CommsChannel = apps.get_model("response", "CommsChannel")
    for comms_channel in CommsChannel.objects.all().iterator():
        try:
            channel_name = settings.SLACK_CLIENT.get_channel_name(
                comms_channel.channel_id
            )
            if not channel_name:
                channel_name = "<channel not found>"
        except SlackError as e:
            raise OperationalError(
                f"""Error connecting to the Slack API: {str(e)}
⚠️  This migration requires access to the Slack API to set CommsChannel.channel_name on existing CommsChannel object. Please make sure that the SLACK_TOKEN environment variable is set to a valid value."""
            )

        comms_channel.channel_name = channel_name
        comms_channel.save()


class Migration(migrations.Migration):

    dependencies = [("response", "0008_externaluser_email")]

    operations = [
        migrations.AddField(
            model_name="commschannel",
            name="channel_name",
            field=models.CharField(null=True, max_length=80),
            preserve_default=False,
        ),
        migrations.RunPython(set_comms_channel_names, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="commschannel",
            name="channel_name",
            field=models.CharField(null=False, max_length=80),
            preserve_default=False,
        ),
    ]



import json

import emoji_data_python
from rest_framework import serializers

from response.core.models import Action, Event, ExternalUser, Incident, TimelineEvent
from response.slack.models import CommsChannel
from response.slack.reference_utils import slack_to_human_readable


class ExternalUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExternalUser
        fields = ("app_id", "external_id", "display_name", "full_name", "email")


class TimelineEventSerializer(serializers.ModelSerializer):
    metadata = serializers.JSONField(allow_null=True, required=False)
    # Read-only field for displaying human-readable slack references. Updates
    # should be applied to the text field.
    text_ui = serializers.SerializerMethodField()

    class Meta:
        model = TimelineEvent
        fields = ("id", "timestamp", "text", "event_type", "metadata", "text_ui")
        read_only_fields = ("id",)

    def get_text_ui(self, instance):
        text_ui = slack_to_human_readable(instance.text)
        text_ui = emoji_data_python.replace_colons(text_ui)
        return text_ui


class ActionSerializer(serializers.ModelSerializer):
    user = ExternalUserSerializer()
    # Read-only field for displaying human-readable slack references. Updates
    # should be applied to the details field.
    details_ui = serializers.SerializerMethodField()

    class Meta:
        model = Action
        fields = ("id", "details", "done", "user", "details_ui")
        read_only_fields = ("id",)

    def create(self, validated_data):
        user = ExternalUser.objects.get(
            app_id=validated_data["user"]["app_id"],
            display_name=validated_data["user"]["display_name"],
            external_id=validated_data["user"]["external_id"],
            full_name=validated_data["user"]["full_name"],
        )
        validated_data["user"] = user
        return Action.objects.create(**validated_data)

    def update(self, instance, validated_data):
        if "user" in validated_data:
            instance.user = ExternalUser.objects.get(
                app_id=validated_data["user"]["app_id"],
                display_name=validated_data["user"]["display_name"],
                external_id=validated_data["user"]["external_id"],
                full_name=validated_data["user"]["full_name"],
            )
        instance.details = validated_data.get("details", instance.details)
        instance.done = validated_data.get("done", instance.done)
        instance.save()
        return instance

    def get_details_ui(self, instance):
        details_ui = slack_to_human_readable(instance.details)
        details_ui = emoji_data_python.replace_colons(details_ui)
        return details_ui


class CommsChannelSerializer(serializers.ModelSerializer):
    class Meta:
        model = CommsChannel
        fields = ("channel_id", "channel_name")


class IncidentSerializer(serializers.ModelSerializer):
    reporter = ExternalUserSerializer(read_only=True)
    lead = ExternalUserSerializer()
    comms_channel = CommsChannelSerializer(read_only=True)
    action_items = ActionSerializer(read_only=True, many=True)

    # This ensures we can't unset severity
    # https://www.django-rest-framework.org/api-guide/fields/#required
    # `required = False` means the field doesn't have to be included when the json request is
    # deserialised (including creation), and so it remains unchanged (if None, it remains None).
    # `allow_null` is set to False by default so we still demand a value is given _if_ it's sent in the json.
    severity = serializers.CharField(required=False)

    class Meta:
        model = Incident
        fields = (
            "action_items",
            "comms_channel",
            "end_time",
            "impact",
            "is_closed",
            "lead",
            "id",
            "report",
            "report_time",
            "report_only",
            "private",
            "reporter",
            "severity",
            "start_time",
            "summary",
        )

    def update(self, instance, validated_data):
        instance.end_time = validated_data.get("end_time", instance.end_time)
        instance.impact = validated_data.get("impact", instance.impact)

        new_lead = validated_data.get("lead", None)
        if new_lead:
            instance.lead = ExternalUser.objects.get(
                display_name=new_lead["display_name"],
                external_id=new_lead["external_id"],
                full_name=new_lead["full_name"],
            )

        instance.report = validated_data.get("report", instance.report)
        instance.start_time = validated_data.get("start_time", instance.start_time)
        instance.summary = validated_data.get("summary", instance.summary)
        instance.severity = validated_data.get("severity", instance.severity)

        # we only allow setting private to true, we don't want to allow private incidents becoming public
        if not instance.private:
            instance.private = validated_data.get("private", instance.private)

        instance.save()
        return instance


class EventSerializer(serializers.ModelSerializer):
    payload = serializers.SerializerMethodField()

    class Meta:
        model = Event
        fields = ("id", "timestamp", "event_type", "payload")
        read_only_fields = ("id", "timestamp", "event_type", "payload")

    def get_payload(self, instance):
        return json.loads(instance.payload)



import bleach
import bleach_whitelist
from django.conf import settings


def sanitize(string):
    # bleach doesn't handle None so let's not pass it
    if string and getattr(settings, "RESPONSE_SANITIZE_USER_INPUT", True):
        return bleach.clean(
            string,
            tags=bleach_whitelist.markdown_tags,
            attributes=bleach_whitelist.markdown_attrs,
            styles=bleach_whitelist.all_styles,
        )

    return string



import logging
from datetime import datetime

from django.conf import settings
from django.db.models.signals import post_save
from django.utils.module_loading import import_string

from response.core.models import Action, Event, Incident
from response.core.serializers import ActionSerializer, IncidentSerializer

logger = logging.getLogger(__name__)


class ActionEventHandler:
    @staticmethod
    def handle(sender, instance: Action, **kwargs):
        if instance.incident.private:
            logger.info(f"Skipping action post_save for private incident: {instance}")
            return
        logger.info(f"Handling post_save for action: {instance}")

        event = Event()
        event.event_type = Event.ACTION_EVENT_TYPE

        # Event payload should be a dict for serializing to JSON.
        event.payload = ActionSerializer(instance).data
        event.payload["incident_id"] = instance.incident.pk
        if "details_ui" in event.payload:
            del event.payload["details_ui"]

        event.timestamp = datetime.now(tz=None)
        event.save()


class IncidentEventHandler:
    @staticmethod
    def handle(sender, instance: Incident, **kwargs):
        if instance.private:
            logger.info(f"Skipping incident post_save for private incident: {instance}")
            return
        logger.info(f"Handling post_save for incident: {instance}")

        event = Event()
        event.event_type = Event.INCIDENT_EVENT_TYPE

        # Event payload should be a dict for serializing to JSON.
        event.payload = IncidentSerializer(instance).data
        # Actions generate their own events, no need to duplicate them here.
        if "action_items" in event.payload:
            del event.payload["action_items"]

        event.timestamp = datetime.now(tz=None)
        event.save()


if hasattr(settings, "ACTION_EVENT_HANDLER_CLASS"):
    cls = import_string(settings.ACTION_EVENT_HANDLER_CLASS)
    post_save.connect(cls.handle, sender=Action)
else:
    post_save.connect(ActionEventHandler.handle, sender=Action)

if hasattr(settings, "INCIDENT_EVENT_HANDLER_CLASS"):
    cls = import_string(settings.INCIDENT_EVENT_HANDLER_CLASS)
    post_save.connect(cls.handle, sender=Incident)
else:
    post_save.connect(IncidentEventHandler.handle, sender=Incident)



# Generated by Django 2.2.3 on 2019-09-04 13:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0002_timelineevent")]

    operations = [
        migrations.AddField(
            model_name="pinnedmessage",
            name="timeline_event",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="response.TimelineEvent",
            ),
        ),
        migrations.AlterField(
            model_name="timelineevent",
            name="event_type",
            field=models.CharField(help_text="Type of event.", max_length=10),
        ),
    ]



from django.conf import settings


class Message:
    def __init__(self):
        self.fallback_text = None
        self.blocks = []

    def set_fallback_text(self, text):
        self.fallback_text = text

    def add_block(self, block):
        self.blocks.append(block)

    def serialize(self):
        serialized = []
        for block in self.blocks:
            serialized.append(block.serialize())
        return serialized

    def send(self, channel, ts=None):
        """
        Build and send the message to the required channel
        """
        return settings.SLACK_CLIENT.send_or_update_message_block(
            channel, blocks=self.serialize(), fallback_text=self.fallback_text, ts=ts
        )


class Block:
    def __init__(self, block_id=None):
        self.block_id = block_id

    def serialize(self):
        raise NotImplementedError


class Section(Block):
    def __init__(self, block_id=None, text=None, accessory=None, fields=None):
        super().__init__(block_id=block_id)
        self.text = text
        self.accessory = accessory
        self.fields = fields

    def add_field(self, field):
        if not self.fields:
            self.fields = []
        self.fields.append(field)

    def serialize(self):
        block = {"type": "section"}

        if not (self.fields or self.text or self.accessory):
            raise ValueError

        if self.block_id:
            block["block_id"] = self.block_id

        if self.fields:
            block["fields"] = [t.serialize() for t in self.fields]
            return block

        if self.text:
            block["text"] = self.text.serialize()

        if self.accessory:
            block["accessory"] = self.accessory.serialize()

        return block


class Actions(Block):
    def __init__(self, block_id=None, elements=None):
        super().__init__(block_id=block_id)
        self.elements = elements

    def add_element(self, element):
        if not self.elements:
            self.elements = []

        self.elements.append(element)

    def serialize(self):
        block = {"type": "actions", "block_id": self.block_id}

        block["elements"] = [e.serialize() for e in self.elements]

        return block


class Divider(Block):
    def serialize(self):
        return {"type": "divider"}


class Confirm(Block):
    def __init__(self, title, text, confirm, deny):
        self.title = title
        self.text = text
        self.confirm = confirm
        self.deny = deny

    def serialize(self):
        return {
            "title": {"type": "plain_text", "text": self.title},
            "text": {"type": "mrkdwn", "text": self.text},
            "confirm": {"type": "plain_text", "text": self.confirm},
            "deny": {"type": "plain_text", "text": self.deny},
        }


class Button:
    def __init__(self, text, action_id, value=None, confirm=None):
        self.text = Text(text=text, text_type="plain_text")
        self.action_id = action_id
        self.value = value
        self.confirm = confirm

    def serialize(self):
        button = {
            "type": "button",
            "text": self.text.serialize(),
            "action_id": self.action_id,
        }

        if self.confirm:
            button["confirm"] = self.confirm.serialize()

        if self.value:
            button["value"] = str(self.value)

        return button


class Text:
    def __init__(self, text, title=None, text_type="mrkdwn", add_new_line=False):
        self.text_type = text_type
        self.text = text
        self.title = title
        self.add_new_line = add_new_line

    def serialize(self):
        text = f"{self.text}\n\u00A0" if self.add_new_line else self.text

        return {
            "type": self.text_type,
            "text": self.text if not self.title else f"*{self.title}*\n{text}",
        }



import logging
from datetime import datetime
from urllib.parse import urljoin

from django.conf import settings
from django.db import models
from django.urls import reverse

from response.core.models.incident import Incident
from response.slack.client import SlackError

logger = logging.getLogger(__name__)


class CommsChannelManager(models.Manager):
    def create_comms_channel(self, incident):
        """
        Creates a comms channel in slack, and saves a reference to it in the DB
        """
        time_string = datetime.now().strftime("%b-%-e-%H-%M-%S")
        name = f"inc-{time_string}".lower()

        try:
            channel_id = settings.SLACK_CLIENT.get_or_create_channel(
                name, auto_unarchive=True
            )
        except SlackError as e:
            logger.error(f"Failed to create comms channel {e}")
            raise

        # If the channel already existed we will need to join it
        # If we are already in the channel as we created it, then this is a No-Op
        try:
            logger.info(f"Joining channel {name}")
            settings.SLACK_CLIENT.join_channel(name)
        except SlackError as e:
            logger.error(f"Failed to join comms channel {e}")
            raise

        try:
            doc_url = urljoin(
                settings.SITE_URL,
                reverse("incident_doc", kwargs={"incident_id": incident.pk}),
            )

            settings.SLACK_CLIENT.set_channel_topic(
                channel_id, f"{incident.report} - {doc_url}"
            )
        except SlackError as e:
            logger.error(f"Failed to set channel topic {e}")
            raise

        comms_channel = self.create(
            incident=incident, channel_id=channel_id, channel_name=name
        )
        return comms_channel


class CommsChannel(models.Model):

    objects = CommsChannelManager()
    incident = models.OneToOneField(Incident, on_delete=models.CASCADE)
    channel_id = models.CharField(max_length=20, null=False)
    channel_name = models.CharField(max_length=80, null=False)

    def post_in_channel(self, message: str):
        settings.SLACK_CLIENT.send_message(self.channel_id, message)

    def rename(self, new_name):
        if new_name:
            try:
                response = settings.SLACK_CLIENT.rename_channel(
                    self.channel_id, new_name
                )
                self.channel_name = response["channel"]["name"]
                self.save()
            except SlackError as e:
                logger.error(
                    f"Failed to rename channel {self.channel_id} to {new_name}. Error: {e}"
                )
                raise e
        else:
            logger.info(f"Attempted to rename channel to nothing. No action take.")

    def __str__(self):
        return self.incident.report



from datetime import datetime, timedelta

from rest_framework import pagination, viewsets

from response.core import serializers
from response.core.models.action import Action
from response.core.models.event import Event
from response.core.models.incident import Incident
from response.core.models.timeline import TimelineEvent
from response.core.models.user_external import ExternalUser


class ExternalUserViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = ExternalUser.objects.all()
    serializer_class = serializers.ExternalUserSerializer


class ActionViewSet(viewsets.ModelViewSet):
    # ViewSets define the view behavior.
    queryset = Action.objects.all()
    serializer_class = serializers.ActionSerializer


class IncidentViewSet(viewsets.ModelViewSet):
    """
    Allows getting a list of Incidents (sorted by report time from newest to
    oldest), and updating existing ones.

    Note that Incidents can only be created via the Slack workflow.
    """

    queryset = Incident.objects.order_by("-report_time")

    serializer_class = serializers.IncidentSerializer
    pagination_class = pagination.LimitOffsetPagination


class IncidentActionViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.ActionSerializer

    def get_queryset(self):
        incident_pk = self.kwargs["incident_pk"]
        return Action.objects.filter(incident_id=incident_pk)

    def perform_create(self, serializer):
        incident_pk = self.kwargs["incident_pk"]
        serializer.save(incident_id=incident_pk)


class IncidentsByMonthViewSet(viewsets.ModelViewSet):
    """
    Allows getting a list of Incidents for a given year/month (sorted by
    report time from newest to oldest), and updating existing ones.

    Note that Incidents can only be created via the Slack workflow.
    """

    queryset = Incident.objects.order_by("-report_time")

    serializer_class = serializers.IncidentSerializer
    pagination_class = pagination.LimitOffsetPagination

    def list(self, request, year, month):
        incidents = self.queryset.filter(
            report_time__year=year, report_time__month=month
        )
        page = self.paginate_queryset(incidents)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)


class IncidentTimelineEventViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.TimelineEventSerializer

    def get_queryset(self):
        incident_pk = self.kwargs["incident_pk"]
        return TimelineEvent.objects.filter(incident_id=incident_pk)

    def perform_create(self, serializer):
        incident_pk = self.kwargs["incident_pk"]
        serializer.save(incident_id=incident_pk)


class EventsViewSet(viewsets.ModelViewSet):
    queryset = Event.objects.all()
    serializer_class = serializers.EventSerializer

    def get_queryset(self):
        from_ts = self.request.query_params.get(
            "from", datetime.now(tz=None) - timedelta(days=1)
        )
        to_ts = self.request.query_params.get("to", datetime.now(tz=None))
        return Event.objects.filter(timestamp__range=(from_ts, to_ts))



import json
import time
from datetime import datetime
from unittest import mock
from unittest.mock import ANY

import pytest
from django.urls import reverse

from response.core.models import ExternalUser, Incident


def test_slash_command_invalid_signature(client):
    timestamp = str(int(datetime.now().timestamp()))
    signature = "v0=0d055ef102d1d89ab8f233d6c345bc0fad85116fdf616cc541e3fb7584617d83"

    r = client.post(
        reverse("slash_command"),
        data=b"",
        content_type="application/x-www-form-urlencoded",
        HTTP_X_SLACK_REQUEST_TIMESTAMP=timestamp,
        HTTP_X_SLACK_SIGNATURE=signature,
    )

    assert r.status_code == 403


def test_slash_command_invokes_dialog(post_from_slack_api, mock_slack):
    data = {"user_id": "U123", "trigger_id": "foo"}
    r = post_from_slack_api("slash_command", data)

    assert r.status_code == 200
    mock_slack.dialog_open.assert_called_once_with(dialog=ANY, trigger_id="foo")


@pytest.mark.django_db(transaction=True)
def test_submit_dialog_creates_incident(post_from_slack_api, mock_slack):
    with mock.patch(
        "response.slack.dialog_handlers.get_user_profile",
        return_value={"name": "Opsy McOpsface"},
    ):

        mock_slack.send_or_update_message_block.return_value = {"ts": "123"}

        summary = "testing dialog submission"
        data = {
            "payload": json.dumps(
                {
                    "type": "dialog_submission",
                    "callback_id": "incident-report-dialog",
                    "user": {"id": "U123"},
                    "channel": {"id": "channel-posted-from"},
                    "response_url": "https://fake-response-url",
                    "submission": {
                        "report": "",
                        "summary": summary,
                        "impact": "",
                        "lead": "U123",
                        "severity": "",
                        "incident_type": "live",
                    },
                    "state": "foo",
                }
            )
        }
        r = post_from_slack_api("action", data)

        assert r.status_code == 200

        # Check if incident has been created

        start_time = datetime.now()
        timeout_secs = 2
        backoff = 0.1
        q = Incident.objects.filter(summary=summary)

        while True:
            d = datetime.now() - start_time
            if d.total_seconds() > timeout_secs:
                pytest.fail(f"waited {timeout_secs}s for condition")
                return
            time.sleep(backoff)

            if q.exists():
                break

        # Check that headline post got created
        mock_slack.send_or_update_message_block.assert_called_with(
            "incident-channel-id", blocks=ANY, fallback_text=ANY, ts="123"
        )

        # Check that we sent an ephemeral message to the reporting user
        mock_slack.send_ephemeral_message.assert_called_with(
            "channel-posted-from", "U123", ANY
        )


@pytest.mark.django_db(transaction=True)
def test_submit_dialog_without_incident_type(post_from_slack_api, mock_slack):
    with mock.patch(
        "response.slack.dialog_handlers.get_user_profile",
        return_value={"name": "Opsy McOpsface"},
    ):

        mock_slack.send_or_update_message_block.return_value = {"ts": "123"}

        summary = "testing dialog submission"
        data = {
            "payload": json.dumps(
                {
                    "type": "dialog_submission",
                    "callback_id": "incident-report-dialog",
                    "user": {"id": "U123"},
                    "channel": {"id": "channel-posted-from"},
                    "response_url": "https://fake-response-url",
                    "submission": {
                        "report": "",
                        "summary": summary,
                        "impact": "",
                        "lead": "U123",
                        "severity": "",
                    },
                    "state": "foo",
                }
            )
        }
        r = post_from_slack_api("action", data)

        assert r.status_code == 200

        # Check if incident has been created

        start_time = datetime.now()
        timeout_secs = 2
        backoff = 0.1
        q = Incident.objects.filter(summary=summary)

        while True:
            d = datetime.now() - start_time
            if d.total_seconds() > timeout_secs:
                pytest.fail(f"waited {timeout_secs}s for condition")
                return
            time.sleep(backoff)

            if q.exists():
                break

        # Check that headline post got created
        mock_slack.send_or_update_message_block.assert_called_with(
            "incident-channel-id", blocks=ANY, fallback_text=ANY, ts="123"
        )

        # Check that we sent an ephemeral message to the reporting user
        mock_slack.send_ephemeral_message.assert_called_with(
            "channel-posted-from", "U123", ANY
        )


@pytest.mark.django_db(transaction=True)
def test_edit_incident(post_from_slack_api, mock_slack):

    mock_slack.send_or_update_message_block.return_value = {"ts": "123"}
    mock_slack.get_user_profile.return_value = {"ts": "123", "name": "Opsy McOpsface"}

    user = ExternalUser.objects.get_or_create(
        app_id="slack", external_id="U123", display_name="Opsy McOpsface"
    )[0]

    incident = Incident.objects.create_incident(
        report="Something happened",
        reporter=user,
        report_time=datetime.now(),
        report_only=False,
        private=False,
        summary="Testing editing incidents - before",
        impact="Lots",
        lead=user,
        severity="1",
    )

    newsummary = "Testing editing incidents - after"
    data = {
        "payload": json.dumps(
            {
                "type": "dialog_submission",
                "callback_id": "incident-edit-dialog",
                "user": {"id": "U123"},
                "channel": {"id": "D123"},
                "response_url": "https://fake-response-url",
                "submission": {
                    "report": "",
                    "summary": newsummary,
                    "impact": "",
                    "lead": "U123",
                    "severity": "1",
                    "incident_type": "",
                },
                "state": incident.id,
            }
        )
    }
    r = post_from_slack_api("action", data)

    assert r.status_code == 200

    start_time = datetime.now()
    timeout_secs = 2
    backoff = 0.2
    i = Incident.objects.filter(summary=newsummary)
    while True:
        d = datetime.now() - start_time
        if d.total_seconds() > timeout_secs:
            pytest.fail(f"waited {timeout_secs}s for condition")
            return
        time.sleep(backoff)

        if i.exists():
            break

    # Assert that the headline post gets updated
    mock_slack.send_or_update_message_block.assert_called_with(
        "incident-channel-id", blocks=ANY, fallback_text=ANY, ts="123"
    )


@pytest.mark.django_db(transaction=True)
def test_edit_private_incident(post_from_slack_api, mock_slack):
    mock_slack.send_or_update_message_block.return_value = {"ts": "123"}
    mock_slack.get_user_profile.return_value = {"ts": "123", "name": "Opsy McOpsface"}

    user = ExternalUser.objects.get_or_create(
        app_id="slack", external_id="U123", display_name="Opsy McOpsface"
    )[0]

    incident = Incident.objects.create_incident(
        report="Something happened",
        reporter=user,
        report_time=datetime.now(),
        report_only=False,
        private=True,
        summary="Testing editing incidents - before",
        impact="Lots",
        lead=user,
        severity="1",
    )

    newsummary = "Testing editing incidents - after"
    data = {
        "payload": json.dumps(
            {
                "type": "dialog_submission",
                "callback_id": "incident-edit-dialog",
                "user": {"id": "U123"},
                "channel": {"id": "D123"},
                "response_url": "https://fake-response-url",
                "submission": {
                    "report": "",
                    "summary": newsummary,
                    "impact": "",
                    "lead": "U123",
                    "severity": "1",
                    "incident_type": "",
                },
                "state": incident.id,
            }
        )
    }
    r = post_from_slack_api("action", data)

    assert r.status_code == 200

    start_time = datetime.now()
    timeout_secs = 2
    backoff = 0.2
    # we assert that the incident is still private, even though the update omitted the property
    i = Incident.objects.filter(summary=newsummary, private=True)
    while True:
        d = datetime.now() - start_time
        if d.total_seconds() > timeout_secs:
            pytest.fail(f"waited {timeout_secs}s for condition")
            return
        time.sleep(backoff)

        if i.exists():
            break

    # Assert that the headline post gets updated
    mock_slack.send_or_update_message_block.assert_called_with(
        "incident-channel-id", blocks=ANY, fallback_text=ANY, ts="123"
    )



import re

from response.slack import cache


def channel_reference(channel_id):
    if channel_id is None:
        return None
    return f"<#{channel_id}>"


def user_reference(user_id):
    return f"<@{user_id}>"


def reference_to_id(value):
    """take a string containing <@U123ABCD> refs and extract first match"""
    m = re.search(r"<@(U[A-Z0-9]+)>", value)
    return m.group(1) if m else None


def user_ref_to_username(value):
    """takes a <@U123ABCD> style ref and returns an @username"""
    # strip the '<@' and '>'
    user_id = reference_to_id(value.group())
    user_profile = cache.get_user_profile(user_id)
    return "@" + user_profile["name"] or user_id


def slack_to_human_readable(value):
    # replace user references (<@U3231FFD>) with usernames (@chrisevans)
    value = re.sub(r"(<@U[A-Z0-9]+>)", user_ref_to_username, value)
    value = re.sub(r"(<#C[A-Z0-9]+\|([\-a-zA-Z0-9]+)>)", r"#\2", value)
    return value



import logging
from datetime import datetime
from typing import Any

from django.conf import settings

from response.core.models import ExternalUser, Incident
from response.slack.cache import get_user_profile
from response.slack.decorators import dialog_handler
from response.slack.reference_utils import channel_reference
from response.slack.settings import INCIDENT_EDIT_DIALOG, INCIDENT_REPORT_DIALOG

logger = logging.getLogger(__name__)


@dialog_handler(INCIDENT_REPORT_DIALOG)
def report_incident(
    user_id: str, channel_id: str, submission: Any, response_url: str, state: Any
):
    report = submission["report"]
    summary = submission["summary"]
    impact = submission["impact"]
    lead_id = submission["lead"]
    severity = submission["severity"]
    private = True if submission.get("private", False) else False

    if "incident_type" in submission:
        report_only = submission["incident_type"] == "report"
    else:
        report_only = False

    name = get_user_profile(user_id)["name"]
    reporter, _ = ExternalUser.objects.get_or_create_slack(
        external_id=user_id, display_name=name
    )

    lead = None
    if lead_id:
        lead_name = get_user_profile(lead_id)["name"]
        lead, _ = ExternalUser.objects.get_or_create_slack(
            external_id=lead_id, display_name=lead_name
        )

    Incident.objects.create_incident(
        report=report,
        reporter=reporter,
        report_time=datetime.now(),
        report_only=report_only,
        private=private,
        summary=summary,
        impact=impact,
        lead=lead,
        severity=severity,
    )

    if report_only and hasattr(settings, "INCIDENT_REPORT_CHANNEL_ID"):
        incidents_channel_ref = channel_reference(settings.INCIDENT_REPORT_CHANNEL_ID)
    else:
        incidents_channel_ref = channel_reference(settings.INCIDENT_CHANNEL_ID)

    text = (
        f"Thanks for raising the incident 🙏\n\nHead over to {incidents_channel_ref} "
        f"to complete the report and/or help deal with the issue"
    )
    settings.SLACK_CLIENT.send_ephemeral_message(channel_id, user_id, text)


@dialog_handler(INCIDENT_EDIT_DIALOG)
def edit_incident(
    user_id: str, channel_id: str, submission: Any, response_url: str, state: Any
):
    report = submission["report"]
    summary = submission["summary"]
    impact = submission["impact"]
    lead_id = submission["lead"]
    severity = submission["severity"]
    private = True if submission.get("private", False) else False

    lead = None
    if lead_id:
        lead_name = get_user_profile(lead_id)["name"]
        lead, _ = ExternalUser.objects.get_or_create_slack(
            external_id=lead_id, display_name=lead_name
        )

    try:
        incident = Incident.objects.get(pk=state)

        if not severity and incident.severity:
            raise Exception("Cannot unset severity")

        if private:
            # making an incident private is irreversible
            incident.private = private

        # deliberately update in this way the post_save signal gets sent
        # (required for the headline post to auto update)
        incident.report = report
        incident.summary = summary
        incident.impact = impact
        incident.lead = lead
        incident.severity = severity
        incident.save()

    except Incident.DoesNotExist:
        logger.error(f"No incident found for pk {state}")



import pytest
from rest_framework.test import APIRequestFactory

from tests.factories import ExternalUserFactory


@pytest.fixture
def arf():
    return APIRequestFactory()


@pytest.fixture()
def api_user(transactional_db):
    e = ExternalUserFactory()
    return e.owner



from datetime import datetime

from django.db import models

from response.core.models import ExternalUser, Incident, TimelineEvent
from response.core.serializers import ExternalUserSerializer
from response.slack.cache import get_user_profile


class PinnedMessageManager(models.Manager):
    def add_pin(self, incident, message_ts, author_id, text):
        name = get_user_profile(author_id)["name"]
        author, _ = ExternalUser.objects.get_or_create_slack(
            external_id=author_id, display_name=name
        )
        timestamp = datetime.fromtimestamp(float(message_ts))

        user_data = ExternalUserSerializer(author).data

        timeline_event = TimelineEvent(
            incident=incident,
            timestamp=timestamp,
            text=text,
            event_type="slack_pin",
            metadata={
                "author": user_data,
                "message_ts": message_ts,
                "channel_id": incident.comms_channel().channel_id,
            },
        )
        timeline_event.save()

        PinnedMessage.objects.get_or_create(
            incident=incident,
            message_ts=message_ts,
            defaults={
                "author": author,
                "text": text,
                "timestamp": timestamp,
                "timeline_event": timeline_event,
            },
        )

    def remove_pin(self, incident, message_ts):
        timestamp = datetime.fromtimestamp(float(message_ts))
        TimelineEvent.objects.filter(incident=incident, timestamp=timestamp).delete()
        PinnedMessage.objects.filter(incident=incident, message_ts=message_ts).delete()


class PinnedMessage(models.Model):
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    author = models.ForeignKey(
        ExternalUser, on_delete=models.PROTECT, blank=False, null=False
    )
    message_ts = models.CharField(max_length=50, blank=False, null=False)
    text = models.TextField()
    timestamp = models.DateTimeField()

    objects = PinnedMessageManager()
    timeline_event = models.ForeignKey(
        TimelineEvent, on_delete=models.CASCADE, null=True
    )

    def __str__(self):
        return f"{self.text}"



# Generated by Django 2.2.3 on 2019-09-04 17:04

import jsonfield.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("response", "0003_auto_20190904_1349")]

    operations = [
        migrations.AlterField(
            model_name="timelineevent",
            name="metadata",
            field=jsonfield.fields.JSONField(
                help_text="Additional fields that can be added by other event types",
                null=True,
            ),
        )
    ]



from django.contrib import admin

from response.core.models import Action, Event, ExternalUser, Incident

admin.site.register(Action)
admin.site.register(Event)
admin.site.register(Incident)
admin.site.register(ExternalUser)



import random

import factory
from faker import Factory

from response.core.models import TimelineEvent

faker = Factory.create()


class TimelineEventFactory(factory.DjangoModelFactory):
    class Meta:
        model = TimelineEvent

    incident = factory.SubFactory("tests.factories.IncidentFactory")

    timestamp = factory.LazyFunction(
        lambda: faker.date_time_between(start_date="-3d", end_date="now", tzinfo=None)
    )
    text = factory.LazyFunction(
        lambda: faker.paragraph(nb_sentences=3, variable_nb_sentences=True)
    )
    event_type = factory.LazyFunction(
        lambda: random.choice(["text", "slack_pin", "incident_update"])
    )

    @factory.lazy_attribute
    def metadata(self):
        if self.event_type == "text":
            return None
        elif self.event_type == "slack_pin":
            return {
                "author": {"display_name": "foo"},
                "message_ts": "12345679.0",
                "channel_id": "C12345",
            }
        elif self.event_type == "incident_update":
            return {
                "update_type": "summary",
                "old_value": faker.paragraph(
                    nb_sentences=3, variable_nb_sentences=True
                ),
                "new_value": faker.paragraph(
                    nb_sentences=3, variable_nb_sentences=True
                ),
            }



# Generated by Django 2.2.3 on 2019-09-27 13:39

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("response", "0010_incident_report_only"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="externaluser", unique_together={("owner", "app_id", "external_id")}
        )
    ]



# Generated by Django 2.2.3 on 2019-09-06 14:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0004_auto_20190904_1704")]

    operations = [
        migrations.AlterField(
            model_name="timelineevent",
            name="event_type",
            field=models.CharField(help_text="Type of event.", max_length=30),
        )
    ]



from datetime import datetime

from django.db import models
from jsonfield import JSONField

from response.core.models.incident import Incident
from response.core.util import sanitize

EVENT_TYPES = (
    ("text", "Freeform text field"),
    ("slack_pin", "A message pinned from Slack"),
)


class TimelineEvent(models.Model):

    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    timestamp = models.DateTimeField(
        null=False, default=datetime.now, help_text="Time of when this event occurred."
    )
    text = models.TextField(help_text="Freeform text describing this event")
    event_type = models.CharField(max_length=30, help_text="Type of event.")
    metadata = JSONField(
        help_text="Additional fields that can be added by other event types", null=True
    )

    def save(self, *args, **kwargs):
        self.text = sanitize(self.text)
        super(TimelineEvent, self).save(*args, **kwargs)


def add_incident_update_event(
    incident, update_type, old_value, new_value, text, timestamp=None
):

    if not timestamp:
        timestamp = datetime.now()

    timeline_event = TimelineEvent(
        incident=incident,
        timestamp=timestamp,
        text=text,
        event_type="incident_update",
        metadata={
            "update_type": update_type,
            "old_value": old_value,
            "new_value": new_value,
        },
    )
    timeline_event.save()



from unittest import mock

import pytest

from response.core.util import sanitize


def test_sanitise_handles_none():
    try:
        sanitize(None)
    except TypeError:
        pytest.fail("Unexpected TypeError handling None")


def test_sanitise_sends_sends_to_bleach():
    with mock.patch("bleach.clean", return_value="good text string"):
        actual = sanitize("bad text string")
        assert actual == "good text string"



import pytest

from response.slack.event_handlers import handle_channel_rename
from tests.factories import IncidentFactory


@pytest.mark.django_db
def test_handle_channel_rename(mock_slack):
    incident = IncidentFactory.create()

    handle_channel_rename(incident, {"channel": {"name": "new-channel-name"}})
    assert incident.comms_channel().channel_name == "new-channel-name"



from django.http import Http404, HttpRequest, HttpResponseForbidden
from django.shortcuts import render

from response.core.models import Incident
from response.decorators import response_login_required
from response.slack.models import PinnedMessage, UserStats


@response_login_required
def incident_doc(request: HttpRequest, incident_id: str):
    try:
        incident = Incident.objects.get(pk=incident_id)
    except Incident.DoesNotExist:
        raise Http404("Incident does not exist")

    # private incident details cannot be viewed by anyone
    if incident.private:
        return HttpResponseForbidden()

    events = PinnedMessage.objects.filter(incident=incident).order_by("timestamp")
    user_stats = UserStats.objects.filter(incident=incident).order_by("-message_count")[
        :5
    ]

    return render(
        request,
        template_name="incident_doc.html",
        context={"incident": incident, "events": events, "user_stats": user_stats},
    )



import json
import logging

from django.conf import settings
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from response.core.models.incident import Incident
from response.slack.authentication import slack_authenticate
from response.slack.cache import update_user_cache
from response.slack.decorators import (
    handle_action,
    handle_dialog,
    handle_event,
    handle_notifications,
)
from response.slack.dialog_builder import (
    Dialog,
    SelectFromUsers,
    SelectWithOptions,
    Text,
    TextArea,
)
from response.slack.settings import INCIDENT_REPORT_DIALOG

logger = logging.getLogger(__name__)


@csrf_exempt
@slack_authenticate
def slash_command(request):
    """
    Handles slash commands from slack
    More details here: https://api.slack.com/slash-commands
    Note: The order the elements are specified is the order they
    appear in the slack dialog

    @param request the request from slack containing the slash command
    @return: return a HTTP response to indicate the request was handled
    """

    user_id = request.POST.get("user_id")
    trigger_id = request.POST.get("trigger_id")
    report = request.POST.get("text")

    dialog = Dialog(
        title="Report an Incident",
        submit_label="Report",
        elements=[
            Text(
                label="Report",
                name="report",
                placeholder="What's the tl;dr?",
                value=report,
            )
        ],
    )

    if hasattr(settings, "INCIDENT_REPORT_CHANNEL_ID"):
        dialog.add_element(
            SelectWithOptions(
                [
                    ("Yes - this is a live incident happening right now", "live"),
                    ("No - this is just a report of something that happened", "report"),
                ],
                label="Is this a live incident?",
                name="incident_type",
                optional=False,
            )
        )

    dialog.add_element(
        SelectWithOptions(
            [("Yes (can't be undone)", "yes")],
            label="Is this a private incident?",
            hint="Will sensitive information be shared in incident comms?",
            name="private",
            optional=True,
        )
    )

    dialog.add_element(
        TextArea(
            label="Summary",
            name="summary",
            optional=True,
            placeholder="Can you share any more details?",
        )
    )

    dialog.add_element(
        TextArea(
            label="Impact",
            name="impact",
            optional=True,
            placeholder="Who or what might be affected?",
            hint="Think about affected people, systems, and processes",
        )
    )

    dialog.add_element(SelectFromUsers(label="Lead", name="lead", optional=True))

    dialog.add_element(
        SelectWithOptions(
            [(s.capitalize(), i) for i, s in Incident.SEVERITIES],
            label="Severity",
            name="severity",
            optional=True,
        )
    )

    logger.info(
        f"Handling Slack slash command for user {user_id}, report {report} - opening dialog"
    )

    dialog.send_open_dialog(INCIDENT_REPORT_DIALOG, trigger_id)
    return HttpResponse()


@csrf_exempt
@slack_authenticate
def action(request):
    """
    Handles actions sent from Slack.

    @param request the request from slack containing an action
    @return: return a HTTP response to indicate the request was handled
    """
    payload = json.loads(request.POST["payload"])
    action_type = payload["type"]

    logger.info(f"Handling Slack action of type '{action_type}'")

    if action_type == "dialog_submission":
        handle_dialog.after_response(payload)
    elif action_type == "block_actions":
        handle_action.after_response(payload)
    else:
        logger.error(f"No handler for action type {action_type}")

    return HttpResponse()


@csrf_exempt
@slack_authenticate
def event(request):
    """
    Handles events sent from Slack.

    Details: We can configure our Slack app receive specific events, and doing so tells
    slack to call this endpoint with a JSON payload representing the event.

    See here for reference: https://api.slack.com/events-api

    @param request the request from slack containing an event
    @return: return a HTTP response to indicate the request was handled
    """
    payload = json.loads(request.body)
    action_type = payload["type"]

    logger.info(f"Handling Slack event of type '{action_type}'")

    if action_type == "event_callback":
        handle_event(payload)
    elif action_type == "url_verification":
        # the url_verification event is called when we change the registered event callback url
        # in the Sl ack app configuration.  It expects us to return the challenge token sent in
        # the request
        return HttpResponse(payload["challenge"])

    return HttpResponse()


@csrf_exempt
def cron_minute(request):
    "Handles actions that need to take place every minute"
    handle_notifications()
    return HttpResponse()


@csrf_exempt
def cron_daily(request):
    "Handles actions that need to take place every day"
    update_user_cache()
    return HttpResponse()



import datetime
import json
import random

import pytest
from django.urls import reverse
from faker import Faker
from rest_framework.test import force_authenticate

from response import serializers
from response.core.views import IncidentsByMonthViewSet, IncidentViewSet
from response.models import ExternalUser, Incident
from tests.factories import IncidentFactory

faker = Faker()


def assert_incident_response(incident):
    assert incident["report_time"]

    assert "end_time" in incident  # end_time can be null for open incidents
    assert "is_closed" in incident  # this can be false
    assert incident["impact"]
    assert incident["report"]
    assert incident["start_time"]
    assert incident["summary"]
    assert incident["severity"]
    assert incident["comms_channel"]

    reporter = incident["reporter"]
    assert reporter["display_name"]
    assert reporter["external_id"]

    lead = incident["lead"]
    assert lead["display_name"]
    assert lead["external_id"]

    assert incident["action_items"]
    for action in incident["action_items"]:
        assert action["details"]
        assert "done" in action
        assert action["user"]


def test_list_incidents(arf, api_user):
    persisted_incidents = IncidentFactory.create_batch(5)

    req = arf.get(reverse("incident-list"))
    force_authenticate(req, user=api_user)
    response = IncidentViewSet.as_view({"get": "list"})(req)

    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    incidents = content["results"]
    assert len(incidents) == len(
        persisted_incidents
    ), "Didn't get expected number of incidents back"

    for idx, incident in enumerate(incidents):
        assert incident["report_time"]

        # incidents should be in order of newest to oldest
        if idx != len(incidents) - 1:
            assert (
                incident["report_time"] >= incidents[idx + 1]["report_time"]
            ), "Incidents are not in order of newest to oldest by report time"

        assert_incident_response(incident)


def test_list_incidents_by_month(arf, api_user):
    IncidentFactory.create_batch(5)

    today = datetime.date.today()
    if today.month == 1:
        last_month = 12
        year = today.year - 1
    else:
        last_month = today.month - 1
        year = today.year

    req = arf.get(
        reverse(
            "incidents-bymonth-list",
            kwargs={"year": str(year), "month": f"{last_month:02d}"},
        )
    )
    print(req)
    force_authenticate(req, user=api_user)
    response = IncidentsByMonthViewSet.as_view({"get": "list"})(req, year, last_month)

    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    print(content)
    assert "results" in content, "Response didn't have results key"

    for incident in content["results"]:
        assert incident["report_time"]
        report_time = datetime.datetime.strptime(
            incident["report_time"], "%Y-%m-%dT%H:%M:%S"
        )
        assert report_time.month == last_month
        assert report_time.year == year

        assert_incident_response(incident)


@pytest.mark.parametrize(
    "update_key,update_value",
    (
        ("", ""),  # no update
        ("impact", faker.paragraph(nb_sentences=2, variable_nb_sentences=True)),
        ("report", faker.paragraph(nb_sentences=1, variable_nb_sentences=True)),
        ("summary", faker.paragraph(nb_sentences=3, variable_nb_sentences=True)),
        (
            "start_time",
            faker.date_time_between(start_date="-3d", end_date="now", tzinfo=None),
        ),
        (
            "end_time",
            faker.date_time_between(start_date="-3d", end_date="now", tzinfo=None),
        ),
        ("severity", str(random.randint(1, 4))),
    ),
)
def test_update_incident(arf, api_user, update_key, update_value):
    """
    Tests that we can PUT /incidents/<pk> and mutate fields that get saved to
    the DB.
    """
    persisted_incidents = IncidentFactory.create_batch(5)

    incident = persisted_incidents[0]
    serializer = serializers.IncidentSerializer(incident)
    serialized = serializer.data

    updated = serialized
    del updated["reporter"]  # can't update reporter
    if update_key:
        updated[update_key] = update_value

    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)

    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-200 response from API"

    if update_key:
        new_incident = Incident.objects.get(pk=incident.pk)
        assert (
            getattr(new_incident, update_key) == update_value
        ), "Updated value wasn't persisted to the DB"


@pytest.mark.parametrize(
    "update_key,update_value,expected_value",
    (
        (
            "impact",
            "<iframe>this should be escaped</iframe>",
            "&lt;iframe&gt;this should be escaped&lt;/iframe&gt;",
        ),
        (
            "report",
            "<script>alert('certainly shouldnt let this happen')</script>",
            "&lt;script&gt;alert('certainly shouldnt let this happen')&lt;/script&gt;",
        ),
        (
            "summary",
            '<meta http-equiv="refresh content=0;">',
            '&lt;meta http-equiv="refresh content=0;"&gt;',
        ),
    ),
)
def test_update_incident_sanitize_fields(
    arf, api_user, update_key, update_value, expected_value
):
    """
    Tests that we can't inject potentially dangerous HTML/JS into incident
    fields
    """
    persisted_incidents = IncidentFactory.create_batch(5)

    incident = persisted_incidents[0]
    serializer = serializers.IncidentSerializer(incident)
    serialized = serializer.data

    updated = serialized
    del updated["reporter"]  # can't update reporter
    if update_key:
        updated[update_key] = update_value

    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)

    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-200 response from API"

    if update_key:
        new_incident = Incident.objects.get(pk=incident.pk)
        print(getattr(new_incident, update_key))
        assert (
            getattr(new_incident, update_key) == expected_value
        ), "Updated value wasn't persisted to the DB"


def test_update_incident_lead(arf, api_user):
    """
    Tests that we can update the incident lead by name
    """
    persisted_incidents = IncidentFactory.create_batch(5)

    incident = persisted_incidents[0]
    serializer = serializers.IncidentSerializer(incident)
    updated = serializer.data

    users = ExternalUser.objects.all()

    new_lead = users[0]
    while new_lead == incident.lead:
        new_lead = random.choices(users)

    updated["lead"] = serializers.ExternalUserSerializer(new_lead).data
    del updated["reporter"]  # can't update reporter

    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)

    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-200 response from API"

    new_incident = Incident.objects.get(pk=incident.pk)
    assert new_incident.lead == new_lead


def test_cannot_unset_severity(arf, api_user):
    """
    Tests that we cannot unset the incident severity
    """

    incident = IncidentFactory.create()
    serializer = serializers.IncidentSerializer(incident)
    updated = serializer.data

    updated["severity"] = None  # unset severity

    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)

    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)
    print(response.rendered_content)
    assert (
        response.status_code != 200
    ), "Got 200 response from API when we expected an error"


def test_make_incident_private(arf, api_user):
    """
    Tests that we cannot unset the incident severity
    """

    incident = IncidentFactory.create()
    serializer = serializers.IncidentSerializer(incident)
    updated = serializer.data

    updated["private"] = True  # unset severity

    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)

    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)
    print(response.rendered_content)
    assert response.status_code == 200, "Incident update succeeded"

    new_incident = Incident.objects.get(pk=incident.pk)
    assert new_incident.private, "Incident not marked as private"


def test_cannot_make_private_incident_public(arf, api_user):
    """
    Tests that we cannot make private incidents public
    """

    incident = IncidentFactory.create(private=True)
    serializer = serializers.IncidentSerializer(incident)
    updated = serializer.data

    updated["private"] = False
    req = arf.put(
        reverse("incident-detail", kwargs={"pk": incident.pk}), updated, format="json"
    )
    force_authenticate(req, user=api_user)
    response = IncidentViewSet.as_view({"put": "update"})(req, pk=incident.pk)

    print(response.rendered_content)
    assert response.status_code == 200, "Update failed"

    new_incident = Incident.objects.get(pk=incident.pk)
    assert new_incident.private, "Private incident marked as public"


def test_cannot_access_incident_logged_out_if_configured(client, db, settings):
    settings.RESPONSE_LOGIN_REQUIRED = True

    incident = IncidentFactory()

    response = client.get(reverse("incident_doc", args=(incident.pk,)))

    assert response.status_code == 302
    assert response["location"].startswith(settings.LOGIN_URL)


def test_cannot_access_private_incident(client, db, settings):
    settings.RESPONSE_LOGIN_REQUIRED = False

    incident = IncidentFactory(private=True)

    response = client.get(reverse("incident_doc", args=(incident.pk,)))

    assert response.status_code == 403


def test_can_access_incident_logged_out_if_configured(client, db, settings):
    settings.RESPONSE_LOGIN_REQUIRED = False

    incident = IncidentFactory()

    response = client.get(reverse("incident_doc", args=(incident.pk,)))

    assert response.status_code == 200



# Generated by Django 2.2.3 on 2019-09-10 14:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("response", "0006_externaluser_full_name")]

    operations = [migrations.DeleteModel(name="IncidentExtension")]



import logging
import time

import slackclient
from slugify import slugify

logger = logging.getLogger(__name__)


class SlackError(Exception):
    def __init__(self, message, slack_error=None):
        self.message = message
        self.slack_error = slack_error


class SlackClient(object):
    def __init__(
        self,
        api_token,
        max_retry_attempts=10,
        retry_base_backoff_seconds=0.2,
        retryable_errors=None,
    ):
        self.api_token = api_token
        self.client = slackclient.SlackClient(self.api_token)
        self.max_retry_attempts = max_retry_attempts
        self.retry_base_backoff_seconds = retry_base_backoff_seconds
        self.retryable_errors = retryable_errors or ["ratelimited"]

    def api_call(self, api_endpoint, *args, **kwargs):
        logger.info(f"Calling Slack API {api_endpoint}")
        response = self.client.api_call(api_endpoint, *args, **kwargs)
        if not response.get("ok", False):
            error = response.get("error", "<no error given>")

            # Check if we want to back off and retry (but only if this isn't a
            # retry attempt itself)
            if error in self.retryable_errors and not kwargs.get("is_retrying", False):
                # Iterate over retry attempts. We could implement this
                # recursively, but there's a danger of overflowing the stack

                # Start index at 1 so we have a multiplier for backoff
                for i in range(1, self.max_retry_attempts + 1):
                    try:
                        # Increase backoff with every attempt
                        backoff_seconds = self.retry_base_backoff_seconds * i
                        logging.warn(
                            f"Retrying request to {api_endpoint} after error {error}. Backing off {backoff_seconds:.2f}s (attempt {i} of {self.max_retry_attempts})"
                        )

                        time.sleep(backoff_seconds)
                        return self.api_call(
                            api_endpoint, is_retrying=True, *args, **kwargs
                        )
                    except SlackError as exc:
                        if exc.slack_error not in self.retryable_errors:
                            raise exc

            raise SlackError(
                f"Error calling Slack API endpoint '{api_endpoint}': {error}",
                slack_error=error,
            )
        return response

    def users_list(self):
        logger.info(f"Listing Slack users")
        return self.api_call("users.list")

    def get_paginated_users(self, limit=0, cursor=None):
        response = self.api_call("users.list", limit=limit, cursor=cursor)
        return response

    def get_user_id(self, name):
        logger.info(f"Getting user ID for {name}")
        response = self.users_list()
        for user in response["members"]:
            if user["name"] == name:
                return user["id"]
        raise SlackError(f"User '{name}' not found")

    def get_channel_name(self, id_):
        try:
            response = self.api_call("conversations.info", channel=id_)
            return response["channel"]["name"]
        except SlackError as e:
            if e.slack_error == "channel_not_found":
                return None
            raise

    def get_channel_id(self, name, auto_unarchive=False):
        logger.info(f"Getting channel ID for {name}")
        next_cursor = None

        while next_cursor != "":
            response = self.api_call(
                "channels.list",
                exclude_archived=not auto_unarchive,
                exclude_members=True,
                limit=800,
                cursor=next_cursor,
            )

            # see if there's a next_cursor
            try:
                next_cursor = response["response_metadata"]["next_cursor"]
                logger.info(f"get_channel_id - next_cursor == [{next_cursor}]")
            except LookupError:
                logger.error(
                    f"get_channel_id - I guess checking next_cursor in response object didn't work."
                )

            for channel in response["channels"]:
                if channel["name"] == name:
                    if channel["is_archived"] and auto_unarchive:
                        self.unarchive_channel(channel["id"])
                    return channel["id"]

        raise SlackError(f"Channel '{name}' not found")

    def get_usergroup_id(self, group_handle):
        response = self.api_call("usergroups.list")
        if not response.get("ok", False):
            raise SlackError(
                f"Failed to get usergroup with group handle {group_handle} : {response['error']}"
            )
        for group in response["usergroups"]:
            if group["handle"] == group_handle:
                return group["id"]
        return None

    def get_usergroup_users(self, group_id):
        response = self.api_call("usergroups.list", include_users=True)

        if not response.get("ok", False):
            raise SlackError(
                f"Failed to get users from usergroup with id {group_id} : {response['error']}"
            )
        for group in response["usergroups"]:
            if group["id"] == group_id:
                return group["users"]
        return None

    def create_channel(self, name):
        response = self.api_call("channels.create", name=name)
        try:
            return response["channel"]["id"]
        except KeyError:
            raise SlackError(
                f"Got unexpected response from Slack API for channels.create - couldn't find channel.id key"
            )

    def get_or_create_channel(self, channel_name, auto_unarchive=False):
        try:
            return self.create_channel(channel_name)
        except SlackError as e:
            if e.slack_error != "name_taken":
                # some other error, let's propagate it upwards
                raise

            return self.get_channel_id(channel_name, auto_unarchive)

    def set_channel_topic(self, channel_id, channel_topic):
        return self.api_call(
            "channels.setTopic", channel=channel_id, topic=channel_topic
        )

    def unarchive_channel(self, channel_id):
        response = self.api_call("channels.unarchive", channel=channel_id)
        if not response.get("ok", False):
            raise SlackError(f"Couldn't unarchive channel {channel_id}")

    def send_message(self, channel_id, text, attachments=None, thread_ts=None):
        return self.api_call(
            "chat.postMessage",
            as_user=False,
            channel=channel_id,
            text=text,
            attachments=attachments,
            thread_ts=thread_ts,
        )

    def send_ephemeral_message(self, channel_id, user_id, text, attachments=None):
        return self.api_call(
            "chat.postEphemeral",
            as_user=False,
            channel=channel_id,
            text=text,
            user=user_id,
            attachments=attachments,
        )

    def send_or_update_message_block(self, channel_id, blocks, fallback_text, ts=None):
        api_call = "chat.postMessage" if not ts else "chat.update"

        return self.api_call(
            api_call, text=fallback_text, channel=channel_id, ts=ts, blocks=blocks
        )

    def add_reaction(self, reaction, channel_id, thread_ts):
        try:
            return self.api_call(
                "reactions.add", name=reaction, channel=channel_id, timestamp=thread_ts
            )
        except SlackError as e:
            if e.slack_error != "already_reacted":
                raise

    def remove_reaction(self, reaction, channel_id, thread_ts):
        try:
            return self.api_call(
                "reactions.remove",
                name=reaction,
                channel=channel_id,
                timestamp=thread_ts,
            )
        except SlackError as e:
            if e.slack_error != "no_reaction":
                raise

    def get_slack_token_owner(self):
        """
        return the id of the user associated with the current slack token
        """
        try:
            return self.api_call("auth.test")["user_id"]
        except KeyError:
            raise SlackError(
                f"Got unexpected response from Slack API for auth.test - couldn't find user_id key"
            )

    def invite_user_to_channel(self, user_id, channel_id):
        return self.api_call("channels.invite", user=user_id, channel=channel_id)

    def join_channel(self, channel_id):
        return self.api_call("channels.join", name=channel_id)

    def leave_channel(self, channel_id):
        return self.api_call("channels.leave", channel=channel_id)

    def get_user_profile(self, user_id):
        if not user_id:
            return None

        response = self.api_call("users.info", user=user_id)

        return {
            "id": user_id,
            "name": response["user"]["name"],
            "fullname": response["user"]["profile"]["real_name"],
            "email": response["user"]["profile"].get("email", None),
        }

    def get_user_profile_by_email(self, email):
        if not email:
            return None

        response = self.api_call("users.lookupByEmail", email=email)

        return {
            "id": response["user"]["id"],
            "name": response["user"]["name"],
            "fullname": response["user"]["profile"]["real_name"],
            "email": email,
        }

    def rename_channel(self, channel_id, new_name):
        prefix = ""
        if not (new_name.startswith("inc-") or new_name.startswith("#inc-")):
            prefix = "inc-"

        new_name = slugify(f"{prefix}{new_name}", max_length=80)

        return self.api_call("channels.rename", channel=channel_id, name=new_name)

    def dialog_open(self, dialog, trigger_id):
        return self.api_call("dialog.open", trigger_id=trigger_id, dialog=dialog)



import json
from datetime import datetime, timedelta

from django.urls import reverse
from rest_framework.test import force_authenticate

from response.core.signals import ActionEventHandler, IncidentEventHandler
from response.core.views import EventsViewSet
from tests.factories.action import ActionFactory
from tests.factories.event import EventFactory
from tests.factories.incident import IncidentFactory


def test_generate_incident_event(arf, api_user):
    incident = IncidentFactory()
    IncidentEventHandler.handle(None, incident)

    req = arf.get(reverse("event-list"))
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == 1, "Expected one event"


def test_generate_action_event(arf, api_user):
    incident = IncidentFactory()
    action = ActionFactory.create(incident=incident)
    ActionEventHandler.handle(None, action)

    req = arf.get(reverse("event-list"))
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == 1, "Expected one event"


def test_list_events(arf, api_user):
    persisted_events = EventFactory.create_batch(10)

    req = arf.get(reverse("event-list"))
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == len(
        persisted_events
    ), "Didn't get expected number of events back"

    for event in events:
        assert event["timestamp"]
        assert event["event_type"]
        assert event["payload"]["report"]


def test_filter_events(arf, api_user):
    ts = datetime.now() + timedelta(days=30)
    EventFactory(timestamp=ts)

    query_params = {"from": ts - timedelta(hours=1), "to": ts + timedelta(hours=1)}
    req = arf.get(reverse("event-list"), data=query_params)
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == 1, "Expected one event"


def test_omit_private_incident(arf, api_user):
    incident = IncidentFactory(private=True)
    IncidentEventHandler.handle(None, incident)

    req = arf.get(reverse("event-list"))
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == 0, "Expected zero events"


def test_omit_private_action(arf, api_user):
    incident = IncidentFactory(private=True)
    action = ActionFactory.create(incident=incident)
    ActionEventHandler.handle(None, action)

    req = arf.get(reverse("event-list"))
    force_authenticate(req, user=api_user)

    response = EventsViewSet.as_view({"get": "list"})(req)
    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"
    events = content["results"]
    assert len(events) == 0, "Expected zero events"



from .core.serializers import (
    ActionSerializer,
    CommsChannelSerializer,
    EventSerializer,
    ExternalUserSerializer,
    IncidentSerializer,
    TimelineEventSerializer,
)

__all__ = (
    "ActionSerializer",
    "CommsChannelSerializer",
    "EventSerializer",
    "ExternalUserSerializer",
    "IncidentSerializer",
    "TimelineEventSerializer",
)



import logging
from datetime import datetime

from django.conf import settings

from response.core.models.incident import Incident
from response.slack.decorators import ActionContext, action_handler
from response.slack.dialog_builder import (
    Dialog,
    SelectFromUsers,
    SelectWithOptions,
    Text,
    TextArea,
)
from response.slack.models import CommsChannel, HeadlinePost
from response.slack.settings import INCIDENT_EDIT_DIALOG

logger = logging.getLogger(__name__)


@action_handler(HeadlinePost.CLOSE_INCIDENT_BUTTON)
def handle_close_incident(ac: ActionContext):
    ac.incident.end_time = datetime.now()
    ac.incident.save()


@action_handler(HeadlinePost.CREATE_COMMS_CHANNEL_BUTTON)
def handle_create_comms_channel(ac: ActionContext):
    if CommsChannel.objects.filter(incident=ac.incident).exists():
        return

    comms_channel = CommsChannel.objects.create_comms_channel(ac.incident)

    # Invite the bot to the channel
    try:
        settings.SLACK_CLIENT.invite_user_to_channel(
            settings.INCIDENT_BOT_ID, comms_channel.channel_id
        )
    except Exception as ex:
        logger.error(ex)

    # Un-invite the user who owns the Slack token,
    #   otherwise they'll be added to every incident channel
    slack_token_owner = settings.SLACK_CLIENT.get_slack_token_owner()
    if ac.incident.reporter != slack_token_owner:
        settings.SLACK_CLIENT.leave_channel(comms_channel.channel_id)

    # Update the headline post to link to this
    headline_post = HeadlinePost.objects.get(incident=ac.incident)
    headline_post.comms_channel = comms_channel
    headline_post.save()


@action_handler(HeadlinePost.EDIT_INCIDENT_BUTTON)
def handle_edit_incident_button(ac: ActionContext):
    dialog_elements = [
        Text(label="Report", name="report", value=ac.incident.report),
        TextArea(
            label="Summary",
            name="summary",
            value=ac.incident.summary,
            optional=True,
            placeholder="Can you share any more details?",
        ),
        TextArea(
            label="Impact",
            name="impact",
            value=ac.incident.impact,
            optional=True,
            placeholder="Who or what might be affected?",
            hint="Think about affected people, systems, and processes",
        ),
        SelectFromUsers(
            label="Lead",
            name="lead",
            value=ac.incident.lead.external_id if ac.incident.lead else None,
            optional=True,
        ),
        SelectWithOptions(
            [(s.capitalize(), i) for i, s in Incident.SEVERITIES],
            value=ac.incident.severity,
            label="Severity",
            name="severity",
            optional=True,
        ),
    ]

    if not ac.incident.private:
        dialog_elements.append(
            SelectWithOptions(
                [("Yes (cant't be undone)", "yes")],
                value=None,
                label="Make the incident private?",
                name="private",
                optional=True,
            )
        )

    dialog = Dialog(
        title=f"Edit Incident {ac.incident.pk}",
        submit_label="Save",
        state=ac.incident.pk,
        elements=dialog_elements,
    )
    dialog.send_open_dialog(INCIDENT_EDIT_DIALOG, ac.trigger_id)



import logging

from django.conf import settings
from django.db import transaction

from response.core.models import ExternalUser
from response.slack.client import SlackError

logger = logging.getLogger(__name__)


def update_user_cache():
    cursor = None
    while cursor != "":
        response = settings.SLACK_CLIENT.get_paginated_users(limit=200, cursor=cursor)

        users = response["members"]

        logger.info(f"Updating {len(users)} users in the cache")
        with transaction.atomic():
            for user in users:
                ExternalUser.objects.update_or_create_slack(
                    external_id=user["id"],
                    defaults={
                        "display_name": user["profile"]["display_name_normalized"]
                        or user["name"],
                        "full_name": user["profile"]["real_name"] or user["name"],
                        "email": user["profile"].get("email", None),
                    },
                )
        cursor = response["response_metadata"].get("next_cursor")


def get_user_profile(external_id):
    """
    Gets a slack user profile:
        - from the DB cache if available
        - or else from the Slack API
    """
    if not external_id:
        return None

    try:
        external_user = ExternalUser.objects.get(external_id=external_id)
        logger.info(f"Got user {external_id} from DB cache")

        return {
            "id": external_user.external_id,
            "name": external_user.display_name,
            "fullname": external_user.full_name,
            "email": external_user.email,
        }
    except ExternalUser.DoesNotExist:
        # profile from slack
        try:
            user_profile = settings.SLACK_CLIENT.get_user_profile(external_id)
        except SlackError:
            logger.error(f"Failed to get user {external_id} from DB cache or Slack")
            raise

        # store it in the DB
        ExternalUser.objects.get_or_create_slack(
            external_id=user_profile["id"],
            defaults={
                "display_name": user_profile["name"],
                "full_name": user_profile["fullname"],
                "email": user_profile["email"],
            },
        )

        logger.info(f"Got user {external_id} from Slack and cached in DB")
        return user_profile


def get_user_profile_by_email(email):
    """
    Gets a slack user profile:
        - from the DB cache if available
        - or else from the Slack API
    """
    if not email:
        return None

    try:
        external_user = ExternalUser.objects.get(email=email)
        logger.info(f"Got user with email {email} from DB cache")

        return {
            "id": external_user.external_id,
            "name": external_user.display_name,
            "fullname": external_user.full_name,
            "email": external_user.email,
        }
    except ExternalUser.DoesNotExist:
        # profile from slack
        try:
            user_profile = settings.SLACK_CLIENT.get_user_profile_by_email(email)
        except SlackError:
            logger.error(
                f"Failed to get user with email {email} from DB cache or Slack"
            )
            raise

        # store it in the DB
        ExternalUser.objects.get_or_create_slack(
            external_id=user_profile["id"],
            defaults={
                "display_name": user_profile["name"],
                "full_name": user_profile["fullname"],
                "email": user_profile["email"],
            },
        )

        logger.info(f"Got user with email {email} from Slack and cached in DB")
        return user_profile



from unittest import mock

import pytest
import slackclient
from django.conf import settings

from response.slack import client
from tests.slack.slack_payloads import user_by_email


@pytest.fixture
def slack_api_mock(monkeypatch):
    client_mock = mock.Mock(spec=slackclient.SlackClient)
    monkeypatch.setattr(settings.SLACK_CLIENT, "client", client_mock)
    return client_mock


@pytest.fixture
def slack_client(slack_api_mock):
    # We set the backoff ridiculously low so that we're not hanging around
    # waiting for retries in tests
    c = client.SlackClient("test-token", retry_base_backoff_seconds=0.0000001)
    c.client = slack_api_mock
    return c


def test_slack_api_call_success(slack_client, slack_api_mock):
    expected_resp = {"ok": True, "response": "bar"}
    slack_api_mock.api_call.return_value = expected_resp

    resp = slack_client.api_call("test_endpoint", "arg1", kwarg2="foo")
    assert resp == expected_resp
    slack_api_mock.api_call.assert_called_with("test_endpoint", "arg1", kwarg2="foo")


def test_slack_api_call_error(slack_client, slack_api_mock):
    slack_api_mock.api_call.return_value = {"ok": False, "error": "test_error"}

    with pytest.raises(client.SlackError) as e:
        slack_client.api_call("test_endpoint", "arg1", kwarg2="foo")
        assert e.slack_error == "test_error"

    slack_api_mock.api_call.assert_called_with("test_endpoint", "arg1", kwarg2="foo")


def test_slack_backoff_rate_limit_succeeded(slack_client, slack_api_mock):
    expected_resp = {"ok": True, "response": "bar"}
    slack_api_mock.api_call.side_effect = [
        {"ok": False, "error": "ratelimited"},
        {"ok": False, "error": "ratelimited"},
        {"ok": False, "error": "ratelimited"},
        {"ok": False, "error": "ratelimited"},
        {"ok": False, "error": "ratelimited"},
        expected_resp,
    ]

    resp = slack_client.api_call("test_endpoint", "arg1", kwarg2="foo")
    assert resp == expected_resp
    assert slack_api_mock.api_call.call_count == 6


def test_slack_backoff_rate_limit_max_retry_attempts(slack_client, slack_api_mock):
    slack_api_mock.api_call.return_value = {"ok": False, "error": "ratelimited"}

    with pytest.raises(client.SlackError) as e:
        slack_client.api_call("test_endpoint", "arg1", kwarg2="foo")
        assert e.slack_error == "test_error"


def test_get_user_profile_by_email(slack_client, slack_api_mock):
    slack_api_mock.api_call.return_value = user_by_email

    # request a user by email
    user = slack_client.get_user_profile_by_email("spengler@ghostbusters.example.com")

    slack_api_mock.api_call.assert_called_with(
        "users.lookupByEmail", email="spengler@ghostbusters.example.com"
    )

    # check we get back the expected user profile
    assert user["id"] == "W012A3CDE"
    assert user["name"] == "spengler"
    assert user["fullname"] == "Egon Spengler"
    assert user["email"] == "spengler@ghostbusters.example.com"



from datetime import datetime

from response.core.models import Incident
from response.slack.decorators import recurring_notification
from response.slack.models import CommsChannel


@recurring_notification(interval_mins=5, max_notifications=5)
def remind_severity(incident: Incident):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.severity:
            comms_channel.post_in_channel(
                "🌤️ This incident doesn't have a severity. Please set one with `@incident severity ...`"
            )
    except CommsChannel.DoesNotExist:
        pass


@recurring_notification(interval_mins=2, max_notifications=5)
def remind_incident_lead(incident: Incident):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.lead:
            comms_channel.post_in_channel(
                "👩‍🚒 This incident hasn't got a lead. Please set one with `@incident lead ...`"
            )
    except CommsChannel.DoesNotExist:
        pass


@recurring_notification(interval_mins=1440, max_notifications=5)
def remind_close_incident(incident: Incident):

    # Only remind on weekdays (weekday returns an ordinal indexed from 0 on Monday)
    if datetime.now().weekday() in (5, 6):
        return

    # Only remind during the day to prevent alerting people at unsociable hours
    if datetime.now().hour not in range(9, 18):
        return

    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.is_closed():
            user_to_notify = incident.lead or incident.reporter
            comms_channel.post_in_channel(
                f":timer_clock: <@{user_to_notify.external_id}>, this incident has been running a long time."
                " Can it be closed now? Remember to pin important messages in order to create the timeline."
            )
    except CommsChannel.DoesNotExist:
        pass



# Generated by Django 2.2.3 on 2019-11-04 10:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0012_commschannel_onetoone")]

    operations = [
        migrations.AddField(
            model_name="incident",
            name="private",
            field=models.BooleanField(default=False),
        )
    ]



# Generated by Django 2.2.3 on 2019-09-24 15:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0009_commschannel_channel_name")]

    operations = [
        migrations.AddField(
            model_name="incident",
            name="report_only",
            field=models.BooleanField(default=False),
        )
    ]



# Generated by Django 2.2.3 on 2019-11-07 17:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0013_incident_private")]

    operations = [
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                ("event_type", models.CharField(max_length=50)),
                ("payload", models.TextField()),
            ],
        )
    ]



# Generated by Django 2.2.3 on 2019-10-24 14:55

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0011_auto_20190927_1339")]

    operations = [
        migrations.AlterField(
            model_name="commschannel",
            name="incident",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, to="response.Incident"
            ),
        )
    ]



import logging

logger = logging.getLogger(__name__)


SLACK_HEADLINE_POST_ACTION_MAPPINGS = {}


def headline_post_action(order=0, func=None):
    """headline_post_action

    Decorator that allows adding a button/action into the headline post. Should be attached
    to functions that take a HeadlinePost and return a Slack Action type (e.g. Button).

    Actions can be conditional - returning None means that no Action will be added.

    The order parameter determines the relative position the action appears in the post -
    actions will be ordered low-high from left to right. Actions with the same order number
    will be added in the order the decorator is executed, which depends on the order of file imports.

    Usage:
    @headline_post_action(100)
    def edit_incident(headline_post):
        return slack.block_kit.Button(":pencil2: Edit", EDIT_INCIDENT_BUTTON, value=headline_post.incident.pk)

    """

    def _wrapper(fn):
        SLACK_HEADLINE_POST_ACTION_MAPPINGS[
            order
        ] = SLACK_HEADLINE_POST_ACTION_MAPPINGS.get(order, []) + [fn]
        return fn

    if func:
        return _wrapper(func)

    return _wrapper
